@@ -26,62 +26,56 @@
 
 #include "namespaces.hh"
 
-class UnknownRecordContent : public DNSRecordContent
+UnknownRecordContent::UnknownRecordContent(const string& zone) 
 {
-public:
-  UnknownRecordContent(const DNSRecord& dr, PacketReader& pr) 
-    : d_dr(dr)
-  {
-    pr.copyRecord(d_record, dr.d_clen);
+  // parse the input
+  vector<string> parts;
+  stringtok(parts, zone);
+  // we need exactly 3 parts, except if the length field is set to 0 then we only need 2
+  if (parts.size() != 3 && !(parts.size() == 2 && boost::equals(parts.at(1), "0"))) {
+    throw MOADNSException("Unknown record was stored incorrectly, need 3 fields, got " + std::to_string(parts.size()) + ": " + zone);
   }
 
-  UnknownRecordContent(const string& zone) 
-  {
-    // parse the input
-    vector<string> parts;
-    stringtok(parts, zone);
-    if(parts.size()!=3 && !(parts.size()==2 && equals(parts[1],"0")) )
-      throw MOADNSException("Unknown record was stored incorrectly, need 3 fields, got "+std::to_string(parts.size())+": "+zone );
-    const string& relevant=(parts.size() > 2) ? parts[2] : "";
-    unsigned int total=pdns_stou(parts[1]);
-    if(relevant.size() % 2 || relevant.size() / 2 != total)
-      throw MOADNSException((boost::format("invalid unknown record length: size not equal to length field (%d != 2 * %d)") % relevant.size() % total).str());
-    string out;
-    out.reserve(total+1);
-    for(unsigned int n=0; n < total; ++n) {
-      int c;
-      sscanf(relevant.c_str()+2*n, "%02x", &c);
-      out.append(1, (char)c);
+  if (parts.at(0) != "\\#") {
+    throw MOADNSException("Unknown record was stored incorrectly, first part should be '\\#', got '" + parts.at(0) + "'");
+  }
+
+  const string& relevant = (parts.size() > 2) ? parts.at(2) : "";
+  unsigned int total = pdns_stou(parts.at(1));
+  if (relevant.size() % 2 || (relevant.size() / 2) != total) {
+    throw MOADNSException((boost::format("invalid unknown record length: size not equal to length field (%d != 2 * %d)") % relevant.size() % total).str());
+  }
+
+  string out;
+  out.reserve(total + 1);
+
+  for (unsigned int n = 0; n < total; ++n) {
+    int c;
+    if (sscanf(&relevant.at(2*n), "%02x", &c) != 1) {
+      throw MOADNSException("unable to read data at position " + std::to_string(2 * n) + " from unknown record of size " + std::to_string(relevant.size()));
     }
-
-    d_record.insert(d_record.end(), out.begin(), out.end());
+    out.append(1, (char)c);
   }
 
-  string getZoneRepresentation(bool noDot) const override
-  {
-    ostringstream str;
-    str<<"\\# "<<(unsigned int)d_record.size()<<" ";
-    char hex[4];
-    for(size_t n=0; n<d_record.size(); ++n) {
-      snprintf(hex, sizeof(hex), "%02x", d_record.at(n));
-      str << hex;
-    }
-    return str.str();
-  }
+  d_record.insert(d_record.end(), out.begin(), out.end());
+}
 
-  void toPacket(DNSPacketWriter& pw) override
-  {
-    pw.xfrBlob(string(d_record.begin(),d_record.end()));
+string UnknownRecordContent::getZoneRepresentation(bool noDot) const
+{
+  ostringstream str;
+  str<<"\\# "<<(unsigned int)d_record.size()<<" ";
+  char hex[4];
+  for (unsigned char n : d_record) {
+    snprintf(hex, sizeof(hex), "%02x", n);
+    str << hex;
   }
+  return str.str();
+}
 
-  uint16_t getType() const override
-  {
-    return d_dr.d_type;
-  }
-private:
-  DNSRecord d_dr;
-  vector<uint8_t> d_record;
-};
+void UnknownRecordContent::toPacket(DNSPacketWriter& pw)
+{
+  pw.xfrBlob(string(d_record.begin(),d_record.end()));
+}
 
 shared_ptr<DNSRecordContent> DNSRecordContent::deserialize(const DNSName& qname, uint16_t qtype, const string& serialized)
 {
@@ -94,7 +88,7 @@ shared_ptr<DNSRecordContent> DNSRecordContent::deserialize(const DNSName& qname,
 
   /* will look like: dnsheader, 5 bytes, encoded qname, dns record header, serialized data */
 
-  string encoded=qname.toDNSString();
+  const auto& encoded = qname.getStorage();
 
   packet.resize(sizeof(dnsheader) + 5 + encoded.size() + sizeof(struct dnsrecordheader) + serialized.size());
 
@@ -117,6 +111,7 @@ shared_ptr<DNSRecordContent> DNSRecordContent::deserialize(const DNSName& qname,
   if (serialized.size() > 0) {
     memcpy(&packet[pos], serialized.c_str(), serialized.size());
     pos += (uint16_t) serialized.size();
+    (void) pos;
   }
 
   MOADNSParser mdp(false, (char*)&*packet.begin(), (unsigned int)packet.size());
@@ -165,6 +160,12 @@ std::shared_ptr<DNSRecordContent> DNSRecordContent::mastermake(const DNSRecord &
   return i->second(dr, pr);
 }
 
+string DNSRecordContent::upgradeContent(const DNSName& qname, const QType& qtype, const string& content) {
+  // seamless upgrade for previously unsupported but now implemented types.
+  UnknownRecordContent unknown_content(content);
+  shared_ptr<DNSRecordContent> rc = DNSRecordContent::deserialize(qname, qtype.getCode(), unknown_content.serialize(qname));
+  return rc->getZoneRepresentation();
+}
 
 DNSRecordContent::typemap_t& DNSRecordContent::getTypemap()
 {
@@ -212,7 +213,7 @@ DNSResourceRecord DNSResourceRecord::fromWire(const DNSRecord& d) {
   return rr;
 }
 
-void MOADNSParser::init(bool query, const std::string& packet)
+void MOADNSParser::init(bool query, const pdns_string_view& packet)
 {
   if (packet.size() < sizeof(dnsheader))
     throw MOADNSException("Packet shorter than minimal header");
@@ -327,6 +328,20 @@ void MOADNSParser::init(bool query, const std::string& packet)
   }
 }
 
+bool MOADNSParser::hasEDNS() const
+{
+  if (d_header.arcount == 0 || d_answers.empty()) {
+    return false;
+  }
+
+  for (const auto& record : d_answers) {
+    if (record.first.d_place == DNSResourceRecord::ADDITIONAL && record.first.d_type == QType::OPT) {
+      return true;
+    }
+  }
+
+  return false;
+}
 
 void PacketReader::getDnsrecordheader(struct dnsrecordheader &ah)
 {
@@ -366,6 +381,15 @@ void PacketReader::copyRecord(unsigned char* dest, uint16_t len)
   d_pos+=len;
 }
 
+void PacketReader::xfrNodeOrLocatorID(NodeOrLocatorID& ret)
+{
+  if (d_pos + sizeof(ret) > d_content.size()) {
+    throw std::out_of_range("Attempt to read 64 bit value outside of packet");
+  }
+  memcpy(&ret.content, &d_content.at(d_pos), sizeof(ret.content));
+  d_pos += sizeof(ret);
+}
+
 void PacketReader::xfr48BitInt(uint64_t& ret)
 {
   ret=0;
@@ -416,7 +440,7 @@ DNSName PacketReader::getName()
 {
   unsigned int consumed;
   try {
-    DNSName dn((const char*) d_content.data(), d_content.size(), d_pos, true /* uncompress */, 0 /* qtype */, 0 /* qclass */, &consumed, sizeof(dnsheader));
+    DNSName dn((const char*) d_content.data(), d_content.size(), d_pos, true /* uncompress */, nullptr /* qtype */, nullptr /* qclass */, &consumed, sizeof(dnsheader));
     
     d_pos+=consumed;
     return dn;
@@ -435,17 +459,17 @@ static string txtEscape(const string &name)
   string ret;
   char ebuf[5];
 
-  for(string::const_iterator i=name.begin();i!=name.end();++i) {
-    if((unsigned char) *i >= 127 || (unsigned char) *i < 32) {
-      snprintf(ebuf, sizeof(ebuf), "\\%03u", (unsigned char)*i);
+  for(char i : name) {
+    if((unsigned char) i >= 127 || (unsigned char) i < 32) {
+      snprintf(ebuf, sizeof(ebuf), "\\%03u", (unsigned char)i);
       ret += ebuf;
     }
-    else if(*i=='"' || *i=='\\'){
+    else if(i=='"' || i=='\\'){
       ret += '\\';
-      ret += *i;
+      ret += i;
     }
     else
-      ret += *i;
+      ret += i;
   }
   return ret;
 }
@@ -502,23 +526,24 @@ string PacketReader::getUnquotedText(bool lenField)
 }
 
 void PacketReader::xfrBlob(string& blob)
-try
 {
-  if(d_recordlen && !(d_pos == (d_startrecordpos + d_recordlen))) {
-    if (d_pos > (d_startrecordpos + d_recordlen)) {
-      throw std::out_of_range("xfrBlob out of record range");
+  try {
+    if(d_recordlen && !(d_pos == (d_startrecordpos + d_recordlen))) {
+      if (d_pos > (d_startrecordpos + d_recordlen)) {
+        throw std::out_of_range("xfrBlob out of record range");
+      }
+      blob.assign(&d_content.at(d_pos), &d_content.at(d_startrecordpos + d_recordlen - 1 ) + 1);
+    }
+    else {
+      blob.clear();
     }
-    blob.assign(&d_content.at(d_pos), &d_content.at(d_startrecordpos + d_recordlen - 1 ) + 1);
-  }
-  else {
-    blob.clear();
-  }
 
-  d_pos = d_startrecordpos + d_recordlen;
-}
-catch(...)
-{
-  throw std::out_of_range("xfrBlob out of range");
+    d_pos = d_startrecordpos + d_recordlen;
+  }
+  catch(...)
+  {
+    throw std::out_of_range("xfrBlob out of range");
+  }
 }
 
 void PacketReader::xfrBlobNoSpaces(string& blob, int length) {
@@ -541,6 +566,107 @@ void PacketReader::xfrBlob(string& blob, int length)
   }
 }
 
+void PacketReader::xfrSvcParamKeyVals(set<SvcParam> &kvs) {
+  while (d_pos < (d_startrecordpos + d_recordlen)) {
+    if (d_pos + 2 > (d_startrecordpos + d_recordlen)) {
+      throw std::out_of_range("incomplete key");
+    }
+    uint16_t keyInt;
+    xfr16BitInt(keyInt);
+    auto key = static_cast<SvcParam::SvcParamKey>(keyInt);
+    uint16_t len;
+    xfr16BitInt(len);
+    
+    if (d_pos + len > (d_startrecordpos + d_recordlen)) {
+      throw std::out_of_range("record is shorter than SVCB lengthfield implies");
+    }
+
+    switch (key)
+    {
+    case SvcParam::mandatory: {
+      if (len % 2 != 0) {
+        throw std::out_of_range("mandatory SvcParam has invalid length");
+      }
+      if (len == 0) {
+        throw std::out_of_range("empty 'mandatory' values");
+      }
+      std::set<SvcParam::SvcParamKey> paramKeys;
+      size_t stop = d_pos + len;
+      while (d_pos < stop) {
+        uint16_t keyval;
+        xfr16BitInt(keyval);
+        paramKeys.insert(static_cast<SvcParam::SvcParamKey>(keyval));
+      }
+      kvs.insert(SvcParam(key, std::move(paramKeys)));
+      break;
+    }
+    case SvcParam::alpn: {
+      size_t stop = d_pos + len;
+      std::vector<string> alpns;
+      while (d_pos < stop) {
+        string alpn;
+        uint8_t alpnLen = 0;
+        xfr8BitInt(alpnLen);
+        if (alpnLen == 0) {
+          throw std::out_of_range("alpn length of 0");
+        }
+        xfrBlob(alpn, alpnLen);
+        alpns.push_back(alpn);
+      }
+      kvs.insert(SvcParam(key, std::move(alpns)));
+      break;
+    }
+    case SvcParam::no_default_alpn: {
+      if (len != 0) {
+        throw std::out_of_range("invalid length for no-default-alpn");
+      }
+      kvs.insert(SvcParam(key));
+      break;
+    }
+    case SvcParam::port: {
+      if (len != 2) {
+        throw std::out_of_range("invalid length for port");
+      }
+      uint16_t port;
+      xfr16BitInt(port);
+      kvs.insert(SvcParam(key, port));
+      break;
+    }
+    case SvcParam::ipv4hint: /* fall-through */
+    case SvcParam::ipv6hint: {
+      size_t addrLen = (key == SvcParam::ipv4hint ? 4 : 16);
+      if (len % addrLen != 0) {
+        throw std::out_of_range("invalid length for " + SvcParam::keyToString(key));
+      }
+      vector<ComboAddress> addresses;
+      auto stop = d_pos + len;
+      while (d_pos < stop)
+      {
+        ComboAddress addr;
+        xfrCAWithoutPort(key, addr);
+        addresses.push_back(addr);
+      }
+      kvs.insert(SvcParam(key, std::move(addresses)));
+      break;
+    }
+    case SvcParam::echconfig: {
+      std::string blob;
+      blob.reserve(len);
+      xfrBlobNoSpaces(blob, len);
+      kvs.insert(SvcParam(key, blob));
+      break;
+    }
+    default: {
+      std::string blob;
+      blob.reserve(len);
+      xfrBlob(blob, len);
+      kvs.insert(SvcParam(key, blob));
+      break;
+    }
+    }
+  }
+}
+
 
 void PacketReader::xfrHexBlob(string& blob, bool keepReading)
 {
@@ -562,21 +688,21 @@ string simpleCompress(const string& elabel, const string& root)
   vstringtok(parts, label, ".");
   string ret;
   ret.reserve(label.size()+4);
-  for(parts_t::const_iterator i=parts.begin(); i!=parts.end(); ++i) {
-    if(!root.empty() && !strncasecmp(root.c_str(), label.c_str() + i->first, 1 + label.length() - i->first)) { // also match trailing 0, hence '1 +'
+  for(const auto & part : parts) {
+    if(!root.empty() && !strncasecmp(root.c_str(), label.c_str() + part.first, 1 + label.length() - part.first)) { // also match trailing 0, hence '1 +'
       const unsigned char rootptr[2]={0xc0,0x11};
       ret.append((const char *) rootptr, 2);
       return ret;
     }
-    ret.append(1, (char)(i->second - i->first));
-    ret.append(label.c_str() + i->first, i->second - i->first);
+    ret.append(1, (char)(part.second - part.first));
+    ret.append(label.c_str() + part.first, part.second - part.first);
   }
   ret.append(1, (char)0);
   return ret;
 }
 
 // method of operation: silently fail if it doesn't work - we're only trying to be nice, don't fall over on it
-void editDNSPacketTTL(char* packet, size_t length, std::function<uint32_t(uint8_t, uint16_t, uint16_t, uint32_t)> visitor)
+void editDNSPacketTTL(char* packet, size_t length, const std::function<uint32_t(uint8_t, uint16_t, uint16_t, uint32_t)>& visitor)
 {
   if(length < sizeof(dnsheader))
     return;
