@@ -3,6 +3,7 @@
 namespace YaHTTP {
 
   template class AsyncLoader<Request>;
+  template class AsyncLoader<Response>;
 
   bool isspace(char c) {
     return std::isspace(c) != 0;
@@ -172,7 +173,9 @@ namespace YaHTTP {
           buffer.copy(buf, pos);
           buf[pos]=0; // just in case...
           buffer.erase(buffer.begin(), buffer.begin()+pos+1); // remove line from buffer
-          sscanf(buf, "%x", &chunk_size);
+          if (sscanf(buf, "%x", &chunk_size) != 1) {
+            throw ParseError("Unable to parse chunk size");
+          }
           if (chunk_size == 0) { state = 3; break; } // last chunk
         } else {
           int crlf=1;
