@@ -5,6 +5,7 @@
 #include <boost/test/unit_test.hpp>
 
 #include "dnsdist.hh"
+#include "dnsdist-lua.hh"
 #include "dnsdist-lua-ffi.hh"
 #include "dolog.hh"
 
@@ -27,6 +28,9 @@ GlobalStateHolder<NetmaskTree<DynBlock>> g_dynblockNMG;
 GlobalStateHolder<SuffixMatchTree<DynBlock>> g_dynblockSMT;
 #endif /* BENCH_POLICIES */
 
+GlobalStateHolder<pools_t> g_pools;
+std::vector<std::unique_ptr<ClientState>> g_frontends;
+
 /* add stub implementations, we don't want to include the corresponding object files
    and their dependencies */
 
@@ -56,7 +60,7 @@ std::string DOHUnit::getHTTPQueryString() const
   return "";
 }
 
-void DOHUnit::setHTTPResponse(uint16_t statusCode, const std::string& body_, const std::string& contentType_)
+void DOHUnit::setHTTPResponse(uint16_t statusCode, PacketBuffer&& body_, const std::string& contentType_)
 {
 }
 #endif /* HAVE_DNS_OVER_HTTPS */
@@ -76,23 +80,31 @@ bool DNSDistSNMPAgent::sendDNSTrap(const DNSQuestion& dq, const std::string& rea
   return false;
 }
 
+void setLuaNoSideEffect()
+{
+}
+
+DNSAction::Action SpoofAction::operator()(DNSQuestion* dq, std::string* ruleresult) const
+{
+  return DNSAction::Action::None;
+}
+
+string g_outputBuffer;
+
 static DNSQuestion getDQ(const DNSName* providedName = nullptr)
 {
   static const DNSName qname("powerdns.com.");
   static const ComboAddress lc("127.0.0.1:53");
   static const ComboAddress rem("192.0.2.1:42");
   static struct timespec queryRealTime;
-  static struct dnsheader dh;
+  static PacketBuffer packet(sizeof(dnsheader));
 
-  memset(&dh, 0, sizeof(dh));
   uint16_t qtype = QType::A;
   uint16_t qclass = QClass::IN;
-  size_t bufferSize = 0;
-  size_t queryLen = 0;
   bool isTcp = false;
   gettime(&queryRealTime, true);
 
-  DNSQuestion dq(providedName ? providedName : &qname, qtype, qclass, qname.wirelength(), &lc, &rem, &dh, bufferSize, queryLen, isTcp, &queryRealTime);
+  DNSQuestion dq(providedName ? providedName : &qname, qtype, qclass, &lc, &rem, packet, isTcp, &queryRealTime);
   return dq;
 }
 
@@ -122,7 +134,7 @@ static void benchPolicy(const ServerPolicy& pol)
   for (size_t idx = 0; idx < 1000; idx++) {
   for (const auto& name : names) {
     auto dq = getDQ(&name);
-    auto server = getSelectedBackendFromPolicy(pol, servers, dq);
+    auto server = pol.getSelectedBackend(servers, dq);
   }
   }
   cerr<<pol.name<<" took "<<std::to_string(sw.udiff())<<" us for "<<names.size()<<endl;
@@ -149,39 +161,129 @@ BOOST_AUTO_TEST_CASE(test_firstAvailable) {
   servers.push_back({ 1, std::make_shared<DownstreamState>(ComboAddress("192.0.2.1:53")) });
 
   /* servers start as 'down' */
-  auto server = getSelectedBackendFromPolicy(pol, servers, dq);
+  auto server = pol.getSelectedBackend(servers, dq);
   BOOST_CHECK(server == nullptr);
 
   /* mark the server as 'up' */
   servers.at(0).second->setUp();
-  server = getSelectedBackendFromPolicy(pol, servers, dq);
+  server = pol.getSelectedBackend(servers, dq);
   BOOST_CHECK(server != nullptr);
 
   /* add a second server, we should still get the first one */
   servers.push_back({ 2, std::make_shared<DownstreamState>(ComboAddress("192.0.2.2:53")) });
-  server = getSelectedBackendFromPolicy(pol, servers, dq);
+  server = pol.getSelectedBackend(servers, dq);
   BOOST_REQUIRE(server != nullptr);
   BOOST_CHECK(server == servers.at(0).second);
 
   /* mark the first server as 'down', second as 'up' */
   servers.at(0).second->setDown();
   servers.at(1).second->setUp();
-  server = getSelectedBackendFromPolicy(pol, servers, dq);
+  server = pol.getSelectedBackend(servers, dq);
   BOOST_REQUIRE(server != nullptr);
   BOOST_CHECK(server == servers.at(1).second);
 
-  std::vector<DNSName> names;
-  names.reserve(1000);
-  for (size_t idx = 0; idx < 1000; idx++) {
-    names.push_back(DNSName("powerdns-" + std::to_string(idx) + ".com."));
+  benchPolicy(pol);
+}
+
+BOOST_AUTO_TEST_CASE(test_firstAvailableWithOrderAndQPS) {
+  auto dq = getDQ();
+  size_t qpsLimit = 10;
+
+  ServerPolicy pol{"firstAvailable", firstAvailable, false};
+  ServerPolicy::NumberedServerVector servers;
+  servers.push_back({ 1, std::make_shared<DownstreamState>(ComboAddress("192.0.2.1:53")) });
+  servers.push_back({ 2, std::make_shared<DownstreamState>(ComboAddress("192.0.2.2:53")) });
+  /* Second server has a higher order, so most queries should be routed to the first (remember that
+     we need to keep them ordered!).
+     However the first server has a QPS limit at 10 qps, so any query above that should be routed 
+     to the second server. */
+  servers.at(0).second->order = 1;
+  servers.at(1).second->order = 2;
+  servers.at(0).second->qps = QPSLimiter(qpsLimit, qpsLimit);
+  /* mark the servers as 'up' */
+  servers.at(0).second->setUp();
+  servers.at(1).second->setUp();
+
+  /* the first queries under the QPS limit should be
+     sent to the first server */
+  for (size_t idx = 0; idx < qpsLimit; idx++) {
+    auto server = pol.getSelectedBackend(servers, dq);
+    BOOST_REQUIRE(server != nullptr);
+    BOOST_CHECK(server == servers.at(0).second);
+    server->incQueriesCount();
   }
+
+  /* then to the second server */
+  for (size_t idx = 0; idx < 100; idx++) {
+    auto server = pol.getSelectedBackend(servers, dq);
+    BOOST_REQUIRE(server != nullptr);
+    BOOST_CHECK(server == servers.at(1).second);
+    server->incQueriesCount();
+  }
+}
+
+BOOST_AUTO_TEST_CASE(test_roundRobin) {
+  auto dq = getDQ();
+
+  ServerPolicy pol{"roundrobin", roundrobin, false};
+  ServerPolicy::NumberedServerVector servers;
+
+  /* selecting a server on an empty server list */
+  g_roundrobinFailOnNoServer = false;
+  auto server = pol.getSelectedBackend(servers, dq);
+  BOOST_CHECK(server == nullptr);
+
+  servers.push_back({ 1, std::make_shared<DownstreamState>(ComboAddress("192.0.2.1:53")) });
+
+  /* servers start as 'down' but the RR policy returns a server unless g_roundrobinFailOnNoServer is set */
+  g_roundrobinFailOnNoServer = true;
+  server = pol.getSelectedBackend(servers, dq);
+  BOOST_CHECK(server == nullptr);
+  g_roundrobinFailOnNoServer = false;
+  server = pol.getSelectedBackend(servers, dq);
+  BOOST_CHECK(server != nullptr);
+
+  /* mark the server as 'up' */
+  servers.at(0).second->setUp();
+  server = pol.getSelectedBackend(servers, dq);
+  BOOST_CHECK(server != nullptr);
+
+  /* add a second server, we should get the first one then the second one */
+  servers.push_back({ 2, std::make_shared<DownstreamState>(ComboAddress("192.0.2.2:53")) });
+  servers.at(1).second->setUp();
+  server = pol.getSelectedBackend(servers, dq);
+  BOOST_REQUIRE(server != nullptr);
+  BOOST_CHECK(server == servers.at(0).second);
+  server = pol.getSelectedBackend(servers, dq);
+  BOOST_REQUIRE(server != nullptr);
+  BOOST_CHECK(server == servers.at(1).second);
+
+  /* mark the first server as 'down', second as 'up' */
+  servers.at(0).second->setDown();
+  servers.at(1).second->setUp();
+  server = pol.getSelectedBackend(servers, dq);
+  BOOST_REQUIRE(server != nullptr);
+  BOOST_CHECK(server == servers.at(1).second);
+
   std::map<std::shared_ptr<DownstreamState>, uint64_t> serversMap;
-  for (size_t idx = 1; idx <= 10; idx++) {
-    servers.push_back({ idx, std::make_shared<DownstreamState>(ComboAddress("192.0.2." + std::to_string(idx) + ":53")) });
-    serversMap[servers.at(idx - 1).second] = 0;
-    servers.at(idx - 1).second->setUp();
+  /* mark all servers 'up' */
+  for (auto& s : servers) {
+    s.second->setUp();
+    serversMap[s.second] = 0;
   }
 
+  for (size_t idx = 0; idx < 1000; idx++) {
+    server = pol.getSelectedBackend(servers, dq);
+    BOOST_REQUIRE(serversMap.count(server) == 1);
+    ++serversMap[server];
+  }
+  uint64_t total = 0;
+  for (const auto& entry : serversMap) {
+    BOOST_CHECK_EQUAL(entry.second, 1000 / servers.size());
+    total += entry.second;
+  }
+  BOOST_CHECK_EQUAL(total, 1000U);
+
   benchPolicy(pol);
 }
 
@@ -193,24 +295,24 @@ BOOST_AUTO_TEST_CASE(test_leastOutstanding) {
   servers.push_back({ 1, std::make_shared<DownstreamState>(ComboAddress("192.0.2.1:53")) });
 
   /* servers start as 'down' */
-  auto server = getSelectedBackendFromPolicy(pol, servers, dq);
+  auto server = pol.getSelectedBackend(servers, dq);
   BOOST_CHECK(server == nullptr);
 
   /* mark the server as 'up' */
   servers.at(0).second->setUp();
-  server = getSelectedBackendFromPolicy(pol, servers, dq);
+  server = pol.getSelectedBackend(servers, dq);
   BOOST_CHECK(server != nullptr);
 
   /* add a second server, we should still get the first one */
   servers.push_back({ 2, std::make_shared<DownstreamState>(ComboAddress("192.0.2.2:53")) });
-  server = getSelectedBackendFromPolicy(pol, servers, dq);
+  server = pol.getSelectedBackend(servers, dq);
   BOOST_REQUIRE(server != nullptr);
   BOOST_CHECK(server == servers.at(0).second);
 
   /* mark the first server as 'down', second as 'up' */
   servers.at(0).second->setDown();
   servers.at(1).second->setUp();
-  server = getSelectedBackendFromPolicy(pol, servers, dq);
+  server = pol.getSelectedBackend(servers, dq);
   BOOST_REQUIRE(server != nullptr);
   BOOST_CHECK(server == servers.at(1).second);
 
@@ -218,7 +320,7 @@ BOOST_AUTO_TEST_CASE(test_leastOutstanding) {
   servers.at(0).second->setUp();
   servers.at(0).second->outstanding = 42;
   servers.at(1).second->setUp();
-  server = getSelectedBackendFromPolicy(pol, servers, dq);
+  server = pol.getSelectedBackend(servers, dq);
   BOOST_REQUIRE(server != nullptr);
   BOOST_CHECK(server == servers.at(1).second);
 
@@ -240,7 +342,7 @@ BOOST_AUTO_TEST_CASE(test_wrandom) {
   benchPolicy(pol);
 
   for (size_t idx = 0; idx < 1000; idx++) {
-    auto server = getSelectedBackendFromPolicy(pol, servers, dq);
+    auto server = pol.getSelectedBackend(servers, dq);
     BOOST_REQUIRE(serversMap.count(server) == 1);
     ++serversMap[server];
   }
@@ -268,7 +370,7 @@ BOOST_AUTO_TEST_CASE(test_wrandom) {
   servers.at(servers.size()-1).second->weight = 100;
 
   for (size_t idx = 0; idx < 1000; idx++) {
-    auto server = getSelectedBackendFromPolicy(pol, servers, dq);
+    auto server = pol.getSelectedBackend(servers, dq);
     BOOST_REQUIRE(serversMap.count(server) == 1);
     ++serversMap[server];
   }
@@ -307,7 +409,7 @@ BOOST_AUTO_TEST_CASE(test_whashed) {
 
   for (const auto& name : names) {
     auto dq = getDQ(&name);
-    auto server = getSelectedBackendFromPolicy(pol, servers, dq);
+    auto server = pol.getSelectedBackend(servers, dq);
     BOOST_REQUIRE(serversMap.count(server) == 1);
     ++serversMap[server];
   }
@@ -330,9 +432,9 @@ BOOST_AUTO_TEST_CASE(test_whashed) {
   /* request 1000 times the same name, we should go to the same server every time */
   {
     auto dq = getDQ(&names.at(0));
-    auto server = getSelectedBackendFromPolicy(pol, servers, dq);
+    auto server = pol.getSelectedBackend(servers, dq);
     for (size_t idx = 0; idx < 1000; idx++) {
-      BOOST_CHECK(getSelectedBackendFromPolicy(pol, servers, dq) == server);
+      BOOST_CHECK(pol.getSelectedBackend(servers, dq) == server);
     }
   }
 
@@ -346,7 +448,7 @@ BOOST_AUTO_TEST_CASE(test_whashed) {
 
   for (const auto& name : names) {
     auto dq = getDQ(&name);
-    auto server = getSelectedBackendFromPolicy(pol, servers, dq);
+    auto server = pol.getSelectedBackend(servers, dq);
     BOOST_REQUIRE(serversMap.count(server) == 1);
     ++serversMap[server];
   }
@@ -392,7 +494,7 @@ BOOST_AUTO_TEST_CASE(test_chashed) {
 
   for (const auto& name : names) {
     auto dq = getDQ(&name);
-    auto server = getSelectedBackendFromPolicy(pol, servers, dq);
+    auto server = pol.getSelectedBackend(servers, dq);
     BOOST_REQUIRE(serversMap.count(server) == 1);
     ++serversMap[server];
   }
@@ -415,9 +517,9 @@ BOOST_AUTO_TEST_CASE(test_chashed) {
   /* request 1000 times the same name, we should go to the same server every time */
   {
     auto dq = getDQ(&names.at(0));
-    auto server = getSelectedBackendFromPolicy(pol, servers, dq);
+    auto server = pol.getSelectedBackend(servers, dq);
     for (size_t idx = 0; idx < 1000; idx++) {
-      BOOST_CHECK(getSelectedBackendFromPolicy(pol, servers, dq) == server);
+      BOOST_CHECK(pol.getSelectedBackend(servers, dq) == server);
     }
   }
 
@@ -431,7 +533,7 @@ BOOST_AUTO_TEST_CASE(test_chashed) {
 
   for (const auto& name : names) {
     auto dq = getDQ(&name);
-    auto server = getSelectedBackendFromPolicy(pol, servers, dq);
+    auto server = pol.getSelectedBackend(servers, dq);
     BOOST_REQUIRE(serversMap.count(server) == 1);
     ++serversMap[server];
   }
@@ -487,7 +589,7 @@ BOOST_AUTO_TEST_CASE(test_lua) {
 
     for (const auto& name : names) {
       auto dq = getDQ(&name);
-      auto server = getSelectedBackendFromPolicy(pol, servers, dq);
+      auto server = pol.getSelectedBackend(servers, dq);
       BOOST_REQUIRE(serversMap.count(server) == 1);
       ++serversMap[server];
     }
@@ -547,7 +649,7 @@ BOOST_AUTO_TEST_CASE(test_lua_ffi_rr) {
 
     for (const auto& name : names) {
       auto dq = getDQ(&name);
-      auto server = getSelectedBackendFromPolicy(pol, servers, dq);
+      auto server = pol.getSelectedBackend(servers, dq);
       BOOST_REQUIRE(serversMap.count(server) == 1);
       ++serversMap[server];
     }
@@ -604,7 +706,7 @@ BOOST_AUTO_TEST_CASE(test_lua_ffi_hashed) {
 
     for (const auto& name : names) {
       auto dq = getDQ(&name);
-      auto server = getSelectedBackendFromPolicy(pol, servers, dq);
+      auto server = pol.getSelectedBackend(servers, dq);
       BOOST_REQUIRE(serversMap.count(server) == 1);
       ++serversMap[server];
     }
@@ -659,7 +761,7 @@ BOOST_AUTO_TEST_CASE(test_lua_ffi_whashed) {
 
     for (const auto& name : names) {
       auto dq = getDQ(&name);
-      auto server = getSelectedBackendFromPolicy(pol, servers, dq);
+      auto server = pol.getSelectedBackend(servers, dq);
       BOOST_REQUIRE(serversMap.count(server) == 1);
       ++serversMap[server];
     }
@@ -721,7 +823,7 @@ BOOST_AUTO_TEST_CASE(test_lua_ffi_chashed) {
 
     for (const auto& name : names) {
       auto dq = getDQ(&name);
-      auto server = getSelectedBackendFromPolicy(pol, servers, dq);
+      auto server = pol.getSelectedBackend(servers, dq);
       BOOST_REQUIRE(serversMap.count(server) == 1);
       ++serversMap[server];
     }
